#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "freetype/freetype.h"
#include "xwin.h"
#include "ftread.h"

static int initted = 0;
FT_Library ftlib;
FT_Face ftface;

int ascent=0;
int descent=0;
int margin=0;
int spacewidth=0;

static XImage *sample_image;

void
load_font()
{
  if (text_fs)
    XFreeFont(display, text_fs);
  if (!font_is_freetype)
    {
      if (font_is_scalable)
	{
	  char *tmpname = (char *)malloc(strlen(fontfile)+10);
	  sprintf(tmpname, fontfile, fontsize);
	  text_fs = XLoadQueryFont(display, tmpname);
	  free(tmpname);
	}
      else
	text_fs = XLoadQueryFont(display, fontfile);
      ascent = text_fs->ascent;
      descent = text_fs->descent;
      margin = 2;
      spacewidth = string_width(" ");
      return;
    }
  text_fs = 0;

  if (!initted)
    {
      initted = 1;
      FT_Init_FreeType(&ftlib);
      sample_image = XGetImage(display, rootwin, 0, 0, 32, 2, ~0, ZPixmap);
    }
  FT_New_Face (ftlib, fontfile, 0, &ftface);
  FT_Set_Pixel_Sizes (ftface, 0, fontsize);

  ascent = ftface->ascender * fontsize / ftface->units_per_EM;
  descent = ftface->descender * fontsize / ftface->units_per_EM;
  int h = ftface->height * fontsize / ftface->units_per_EM;
  margin = (h - (ascent+descent)) / 2;
  ascent += (h - ascent - descent) / 2;
  descent = h - ascent;
  spacewidth = string_width (" ");
}

int
string_width (char *string)
{
  int maxx=0;
  int penx=0;
  int i;

  if (!font_is_freetype)
    return XTextWidth (text_fs, string, strlen(string));

  FT_GlyphSlot slot = ftface->glyph;

  for (i=0; string[i]; i++)
    {
      FT_Load_Char (ftface, string[i], 0);

      int px = (penx + 32) & ~63;
      FT_Glyph_Metrics *m = &(slot->metrics);
      
      FT_Vector kerning;
      if (i > 0 && 0 == FT_Get_Kerning (ftface,
					FT_Get_Char_Index(ftface, string[i-1]),
					FT_Get_Char_Index(ftface, string[i]), 0, &kerning))
	{
	  penx += kerning.x;
	}

      if (maxx < penx + m->horiBearingX + m->width)
	maxx = penx + m->horiBearingX + m->width;

      penx += slot->advance.x;
    }

  if (maxx < penx)
    maxx = penx;

  return maxx / 64;
}

XImage *
build_string(char *string, int *y_adj)
{
  int minx=0, miny=0, maxx=0, maxy=0;
  int penx=0, peny=0;
  int i;

  if (!font_is_freetype)
    {
      if (y_adj)
	*y_adj = ascent;
      return (XImage *)strdup(string);
    }

  FT_GlyphSlot slot = ftface->glyph;

  for (i=0; string[i]; i++)
    {
      FT_Load_Char (ftface, string[i], 0);

      int px = (penx + 32) & ~63;
      FT_Glyph_Metrics *m = &(slot->metrics);
      
      FT_Vector kerning;
      if (i > 0 && 0 == FT_Get_Kerning (ftface,
					FT_Get_Char_Index(ftface, string[i-1]),
					FT_Get_Char_Index(ftface, string[i]), 0, &kerning))
	{
	  penx += kerning.x;
	  peny += kerning.y;
	}

      if (minx > penx + m->horiBearingX)
	minx = penx + m->horiBearingX;
      if (maxx < penx + m->horiBearingX + m->width)
	maxx = penx + m->horiBearingX + m->width;

      if (miny > m->horiBearingY - m->height)
	miny = m->horiBearingY - m->height;
      if (maxy < m->horiBearingY)
	maxy = m->horiBearingY;

      penx += slot->advance.x;
      peny += slot->advance.y;
    }

#if 0
  if (miny > ftface->size->metrics.descender)
    miny = ftface->size->metrics.descender;
  if (maxy < ftface->size->metrics.ascender)
    maxy = ftface->size->metrics.ascender;
#endif
  if (maxx < penx)
    maxx = penx;

  int w = (maxx - minx)/64 + 2;
  int h = (maxy - miny)/64 + 2;

  if (y_adj)
    *y_adj = (ftface->size->metrics.ascender - maxy) / 64;

  static XImage *rv = 0;
  rv = XCreateImage (display,
		     DefaultVisual(display, screen),
		     sample_image->depth,
		     sample_image->format,
		     0, 0, w, h,
		     sample_image->bitmap_pad,
		     0);
		     
  rv->data = (char *)malloc (rv->bytes_per_line * h + sizeof(int));

  for (penx=0; penx<w; penx++)
    for (peny=0; peny<h; peny++)
      XPutPixel(rv, penx, peny, grays[0]);

  penx = -minx;
  peny = -miny;


  for (i=0; string[i]; i++)
    {
      FT_Load_Char (ftface, string[i], FT_LOAD_RENDER);

      FT_Vector kerning;
      if (i > 0 && 0 == FT_Get_Kerning (ftface,
					FT_Get_Char_Index(ftface, string[i-1]),
					FT_Get_Char_Index(ftface, string[i]), 0, &kerning))
	{
	  penx += kerning.x;
	  peny += kerning.y;
	}

      int left = penx/64 + slot->bitmap_left;
      int top = peny/64 + slot->bitmap_top;
      int dy, dr;
      unsigned char *row = slot->bitmap.buffer;
      int x, y, nrows;

      if (slot->bitmap.pitch > 0)
	{
	  dy = 1;
	  dr = slot->bitmap.pitch;
	  y = top;
	}
      else
	{
	  dy = -1;
	  dr = -slot->bitmap.pitch;
	  y = top - slot->bitmap.rows + 1;
	}
      y = h - y - 1;

      for (nrows = 0; nrows < slot->bitmap.rows; nrows++)
	{
	  for (x=0; x<slot->bitmap.width; x++)
	    XPutPixel(rv, left+x, y, grays[((unsigned char *)row)[x]/16]);

	  y += dy;
	  row += dr;
	}
      penx += slot->advance.x;
      peny += slot->advance.y;
    }

  return rv;
}
